# _*_ coding: utf-8 _*_
"""
Time:     2021/7/22 17:29
Author:   WANG Bingchen
Version:  V 0.1
File:     __init__.py
Describe: 
"""
from .version import __version__
from .model import SMURF
from .model_sym import SMURF as SMURF_SYM

name = 'smurf-imputation'
